#include "q.h"

static char Ebad[] = "something bad happened";
static char Enomem[] = "no memory";

/*
 * File.aux points to a Queue
 * Fid.aux points to a Reader
 * Req.aux points to the next Req
 */

#define min(a, b) ((a) < (b) ? (a) : (b))

/* Respond to all possible requests on this reader */
void
respondall(Reader *rd)
{
	Req *nextr;
	Message *m;
	int len;

	/* cancel requests on dead readers */
	if(rd->pos < 0){
		while(rd->rq){
			nextr = rd->rq->aux;
			respond(rd->rq, "lag");
			rd->rq = nextr;
		}
		return;
	}
	/* try to respond to as many as possible */
	while(rd->rq && (m = peekmsg(rd))){
		len = m->len - rd->offset;
		len = min(len, rd->rq->ifcall.count);
		rd->rq->ofcall.count = len;
		memmove(rd->rq->ofcall.data, (char*)(m+1)+rd->offset, len);
		rd->offset += len;
		if(rd->offset >= m->len){
			releasemsg(getmsg(rd));
			rd->offset = 0;
		}
		nextr = rd->rq->aux;
		respond(rd->rq, nil);
		rd->rq = nextr;
	}
}

/* Enqueue request and respond to all possible */
void
fsread(Req *r)
{
	Req **rp;
	Reader *rd;

	rd = r->fid->aux;
	assert(rd != nil);
	for(rp = &rd->rq; *rp; rp = &(*rp)->aux);
	*rp = r;
	r->aux = nil;

	respondall(rd);
}

/* Put message into queue and respond on all readers */
void
fswrite(Req *r)
{
	Queue *q;
	Reader *rd;

	q = r->fid->file->aux;
	addmsg(q, r->ifcall.data, r->ifcall.count);
	for(rd = q->rd; rd; rd = rd->next)
		respondall(rd);
	respond(r, nil);
}

/* Remove old request from reader */
void
fsflush(Req *r)
{
	Reader *rd;
	Req **rp;

	rd = r->oldreq->fid->aux;
	for(rp = &rd->rq; *rp; rp = &(*rp)->aux)
		if(*rp == r->oldreq){
			*rp = (*rp)->aux;
			respond(r->oldreq, "interrupted");
			break;
		}
	respond(r, nil);
}

void
fscreate(Req *r)
{
	File *f;
	Queue *q;
	Reader *rd;
	int o;

	if(r->ifcall.perm & DMDIR){
		respond(r, "permission denied");
		return;
	}
	if(f = createfile(r->fid->file, r->ifcall.name, r->fid->uid, r->ifcall.perm, nil)){
		r->fid->file = f;

print("making queue %s\n", f->name);
		q = emalloc9p(sizeof(Queue));
		q->wr = 0;
		q->rd = nil;
		r->fid->file->aux = q;
		o = r->ifcall.mode & OMASK;
		if(o == OREAD || o == ORDWR){
			rd = newreader(q);
			r->fid->aux = rd;
		}else
			r->fid->aux = nil;

		r->ofcall.qid = f->qid;
		respond(r, nil);
		return;
	}
	respond(r, Ebad);
}

void
fsopen(Req *r)
{
	Queue *q;
	Reader *rd;
	Fid *fid;
	int o;

	fid = r->fid;
	q = fid->file->aux;
	o = r->ifcall.mode & OMASK;
	if((fid->qid.type & QTDIR) == 0 && (o == OREAD || o == ORDWR)){
		rd = newreader(q);
		fid->aux = rd;
	}else
		fid->aux = nil;
	respond(r, nil);
}

void
fsdestroyfid(Fid *fid)
{
	Reader *rd;
	rd = fid->aux;
	if(rd == nil)
		return;
	assert(rd->q != nil);
	removereader(rd->q, rd);
print("freeing reader on %s\n", fid->file->name);
	free(rd);
}

void
fsdestroyfile(File *f)
{
	Queue *q;
	q = f->aux;
	assert(q->rd == nil);
print("freeing queue %s\n", f->name);
	free(q);
}

Srv fs = {
	.open		fsopen,
	.read		fsread,
	.write		fswrite,
	.flush		fsflush,
	.create		fscreate,
	.destroyfid	fsdestroyfid,
};

void
usage(void)
{
	fprint(2, "usage: repfs [-D] [-s srvname] [-m mtpt]\n");
	exits("usage");
}

void
main(int argc, char **argv)
{
	char *addr = nil;
	char *srvname = nil;
	char *mtpt = "/mnt/tty";
	Qid q;

	fs.tree = alloctree(nil, nil, DMDIR|0777, fsdestroyfile);
	q = fs.tree->root->qid;

	ARGBEGIN{
	case 'D':
		chatty9p++;
		break;
	case 'a':
		addr = EARGF(usage());
		break;
	case 's':
		srvname = EARGF(usage());
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	default:
		usage();
	}ARGEND;

	if(argc)
		usage();

	if(chatty9p)
		fprint(2, "ramsrv.nopipe %d srvname %s mtpt %s\n", fs.nopipe, srvname, mtpt);
	if(addr == nil && srvname == nil && mtpt == nil)
		sysfatal("must specify -a, -s, or -m option");
	if(addr)
		listensrv(&fs, addr);

	if(srvname || mtpt)
		postmountsrv(&fs, srvname, mtpt, MREPL|MCREATE);
	exits(0);
}
